#ifndef MATHTREE_H
#define MATHTREE_H
/*
   Definitions for math expressions
   ================================


   Copyright  (C) 2002-2009 M.Kraemer, D-64291 Darmstadt, Seegartenstr. 32,
                             all rights reserved

   $Log: mathtree.h,v $
   Revision 1.1.2.1  2005/03/15 14:26:00  kraemer
   *** empty log message ***


*/


#include "ulists.h"
#include "m.h"

/*
#define DBL_SKIP DBL_QNAN
#define FLT_SKIP FLT_QNAN
*/
                                      /* masks */
#define MATHTREE_MASKGLOBAL 0xFF000000    /* global (e.g. debug) */
#define MATHTREE_MASKFUNC   0x00FF0000    /* function specific flags */
#define MATHTREE_MASKQUAL   0x0000FFFF    /* for additional qualifiers */
                                      /* global flags */
#define MATHTREE_DEBUG  0x80000000        /* general check flag */
#define MATHTREE_QUIET  0x10000000        /* suppress error messages */
#define MATHTREE_NOMSG  MATHTREE_QUIET

#define MATHTREE_ERROR     0x80000000
#define MATHTREE_WARNING   0x40000000
#define MATHTREE_ERRORMASK 0xF0000000
#define MATHTREE_ERRMASK   MATHTREE_ERRORMASK 

#define MATHTREE_MAXDIM 16
                                      /* valid first chars in variable */
#define ISMATHTREECHARF(c) ( isalpha(c)               || ( ((c) & 127) == 0x5f) || ( (c) == '$' ) || ( (c) == '#' ) )

                                      /* valid chars in variable */
#define ISMATHTREECHAR(c) ( ISMATHTREECHARF(c) || isdigit(c) )
 
 
                                      /* help with value import by string arrays */
struct SMATHTREEVALSTR {
   int ii;                            /* current item index */ 
   int nn;                            /* # of items */ 
   int mm;                            /* # of all items, incl. repetition */ 
   char **pvc;                        /* [nn] ptr to item (reference) */ 
   int   *nvc;                        /* [nn] len of item */
   int   *nvr;                        /* [nn] optional repetition factor,
                                                <0: item str extra allocated
                                         might get updated when used for import
                                      */
};
 
 
                                      /* description of builtin functions */
struct SMATHTREEBUILTIN {
   char *pcOpName; 
   int nArg; 
   unsigned long ulOpCode;
};
                                      /* builtin op codes */
enum {
   MATHTREE_NOP=0,                    /* No-Op, assignment */
   MATHTREE_ADD, MATHTREE_SUB, MATHTREE_MUL, MATHTREE_DIV,
   MATHTREE_POW, MATHTREE_POW2, MATHTREE_POW10,
   MATHTREE_SQRT,                     /* 0x8 */
   MATHTREE_ABS,
   MATHTREE_CEIL, MATHTREE_FLOOR, MATHTREE_MOD, 
   MATHTREE_MAX, MATHTREE_MIN, 
   MATHTREE_RANDOM,                   /* 0xF */
   MATHTREE_EXP,
   MATHTREE_J0, MATHTREE_J1, MATHTREE_GAMMA, MATHTREE_POISSON,
                                      /* 0x15 */  
   MATHTREE_LOG, MATHTREE_LOG10, MATHTREE_LOG2,
                                      /* 0x18 */  
   MATHTREE_SIN,   MATHTREE_COS,   MATHTREE_TAN, MATHTREE_COT,
                                      /* 0x1C */  
   MATHTREE_SIND,   MATHTREE_COSD,   MATHTREE_TAND, MATHTREE_COTD,
                                      /* 0x20 */  
   MATHTREE_ASIN,  MATHTREE_ACOS,  MATHTREE_ATAN, MATHTREE_ATAN2,
                                      /* 0x24 */  
   MATHTREE_ASIND,  MATHTREE_ACOSD,  MATHTREE_ATAND, MATHTREE_ATAN2D,
                                      /* 0x28 */  
   MATHTREE_SINH,  MATHTREE_COSH,  MATHTREE_TANH,
                                      /* 0x2B */  
   MATHTREE_ASINH, MATHTREE_ACOSH, MATHTREE_ATANH,
   MATHTREE_DIFFERENTIAL, MATHTREE_INTEGRAL,
   MATHTREE_MOVE,                     /* assignment */ 
};

                                      /* subscripts for one dimension */
struct SMATHTREESUB {
   char *pcStr;                       /* subscript string (e.g. from declaration) */
   unsigned long ulCnt;               /* # array elements, this dimension */
   long lLB, lHB, lStep;              /* lower and upper bound (C-syntax), stride */
   long lx;                           /* current array index */
   unsigned long ulTensorIndex;       /* for tensor mode */   
   long lxErr;                        /* error index, 1..n */ 
};

/*
*/
                                      /* for constant/scalar values */
union UNMATHTREEVAL { double dd; float rr; long ll; short ii; char bb; char *pc; };
                                       
                                      /* a math tree node */    
struct SMATHTREE {
   struct Node ln;                    /* stem name */
   unsigned long ulFlag;              /* flag bits */
   unsigned long ulFlagInternal;      /* flag bits not to be duplicated */
   char *pcErrChar;                   /* points to character in error */
   long lLevel;                       /* to support evaluation of concatenated mt's */
   long lType;                        /* data type */
   unsigned long ulData;              /* total # of declared array elements */
   unsigned long ulDataImported;      /* special: # of data array elements imported */
   int nDim;                          /* dimension */
   void **ppv;                        /* data array, from dimalloc() */ 
   union UNMATHTREEVAL unVal;         /* value of a scalar/constant */ 
                                      /* declared subscripts */ 
   struct SMATHTREESUB svMTS[ MATHTREE_MAXDIM ];
                                      /* subscripts used for data import */ 
   struct SMATHTREESUB svMTSImport[ MATHTREE_MAXDIM ];
                                      /* for operations: */
   unsigned long ulOpCode;            /* internal op code */
   int nMT;                           /* # of operands */
   struct SMATHTREE *psMT;            /* []: operands */ 
};
                                      /* flag bits */
#define MATHTREE_PLISUB      0x00000001 /* subscripts in PL/I style, 1..n */
#define MATHTREE_DATAREFCOPY 0x00008000 /* data are not alloc()d, so don't free() */ 
                                      /* opcode bits */
#define MATHTREE_NEGATE 0x80000000    /* negate this operand */
#define MATHTREE_MASKOP 0x7FFFFFFF    /* opcode w/o neg bit */ 
                                      /* data types */
enum { MATHTREE_NOTYPE=-1, MATHTREE_BOOL=0, MATHTREE_SHORT, MATHTREE_LONG, MATHTREE_FLOAT, MATHTREE_DOUBLE, MATHTREE_CHAR, };
                                      /* function action flags ... */
                                      /* */ 
#define MATHTREEIMPORTDATA_REFCOPY     0x00010000 
#define MATHTREEFROMSTR_GDSPECIAL      0x00020000 
#define MATHTREEIMPORTVALSTR_TRANSPOSE 0x00040000 

/*


  ! tensor mode for allocation / evaluation
  ! propagate SKIP bit


   _mathtreeCollectUndefined( mtto, mtfr, flag )
                            o collect undefined operands
                              must be initialized before call !
                              no header, only names in psMT, real copy
                         10 x to test
                            o creates duplicate entries, e.g. for
                              "x (y1=y3*2+y3**2) (y3=pol(x;2,4)),l12 (y4=x*4+2),r11"
                            o calls
                              o _mathtreeDupNoData()
                              o _mathtreeCollectUndefined()
                             

   _mathtreeCopyDataSub     ( mt, mtfr, flag )
                            o copy node data, w/ subrange control
                          6 x to test
                            ? sub lb==hb==0: 1 loop pass, OK ?
                            o calls
                              o _mathtreeCopyDataSub()
                   
   _mathtreeDataAlloc       ( mt, flag )
                            o allocate data space
                         4a x to test
                            o tensor mode 
                            o calls
                              o _mathtreeDataAlloc()

   _mathtreeDataAllocQuick  ( mt, length, type, flag )
                            o allocate data space quick, 1dim
                         4b x to test
                            o calls
                              o _mathtreeDataAlloc()

   _mathtreeDataReset       ( mt, flag )
                            o reset data 
                          8 x to test
                            ? follow operand nodes
                            o calls
                              o _mathtreeDataReset

   _mathtreeDup             ( mtto, mtfr, flag )
                            o duplicate a node
                         7b o to test (recursive case)
                            ? recursive optional ?
                            o calls
                              o _mathtreeInit()
                              o _mathtreeFree()
                              o _mathtreeDataAlloc()
                              o _mathtreeCopyDataSub()
                              o _mathtreeDupNoData()
                              o _mathtreeDup()
                             
   _mathtreeDupNoData       ( mtto, mtfr, flag )
                            o duplicate a node's header info
                         7a x to test
                            o calls
                              o _mathtreeInit()
                              o _mathtreeFree()
                             
   _mathtreeEvaluate        ( mt, flag )
                            o evaluate a math tree 
                         15 x to test
                            ? subranges
                            o tensor mode
                            o calls
                              o _mathtreeEvaluate()
                              o _mathtree1Oper()
                              o _mathtree2Oper()
  
   _mathtree1Oper           ( mtto, mtfr, flag )
                            o multidim operations w/ 1 operand 
                         13 x to test
                            o tensor mode
                            ? subranges/stride
                            ? short/long skip
                            o functions
                              o random()
                            ? histograms 
                            o calls
                              o _mathtree1Oper()
   
   _mathtree2Oper           ( mtto, mt1, mt2, flag )
                            o multidim operations w/ 2 operands
                         14 x to test
                            o tensor mode
                            ? histograms 
                            o subranges/stride ?
                            ? short/long skip
                            o calls
                              o _mathtree2Oper()
   
   _mathtreeFree            ( mt, flag )
                            o release resources
                          3 x to test
                            o calls
                              o _mathtreeFree()
                              o _mathtreeInit()

   _mathtreeFromStr         ( mathtree, cstr, nstr, flag )
                            o convert arithmetic string to math tree
                         12 x to test
                            o calls
                              o _mathtreeFree()
                              o _mathtreeInit()
                              o _mathtreeScan() 
                    
   _mathtreeHasSkip         ( mt, flag );
                            o test the SKIP bit, returns TRUE/FALSE 
                            o to do
                            o calls
                   
   _mathtreeImportData      ( mtto, mtfr, flag )
                            o import data arrays
                          9 x to test
                            ? optionally report on "not-found "
                            o calls
                              o _mathtreeImportData()
                              o _mathtreeDataAlloc()
                              o _mathtreeSetImportSub()
                              o _mathtreeCopyDataSub()
                       
   _mathtreeImportValStrSub ( mt, mtvs, flag )
                            o import data from char ptr array w/ length
                         11 x to test
                            o calls

   _mathtreeInit            ( mt, flag )
                            o initialize a node
                          2 x to test
                            o calls
        
   _mathtreeInspect         ( mt, flag )
                            o inspect a node
                          1 x to test
                            o calls
                              o _mathtreeInspect()
        
   _mathtreeSetImportSub    ( mt, mtsub, flag );
                            o reset/import subscripts for data import 
                          5 x to test
                            o calls
                   
   _mathtreeSetSkip         ( mt, flag );
                            o set the SKIP bit 
                            o to do
                            o calls
                   
   _mathtreeStrError        ( error )
                            o return error string
                          0 x to test
                            o calls
                            
        


*/
                                      /* function prototypes */
int _mathtreeCollectUndefined( struct SMATHTREE *, struct SMATHTREE *, unsigned long );
int _mathtreeCopyDataSub     ( struct SMATHTREE *, struct SMATHTREE *, unsigned long );
int _mathtreeDataAlloc       ( struct SMATHTREE *, unsigned long );
int _mathtreeDataAllocQuick  ( struct SMATHTREE *, int, long, unsigned long );
int _mathtreeDataReset       ( struct SMATHTREE *, unsigned long );
int _mathtreeDup        ( struct SMATHTREE *, struct SMATHTREE *, unsigned long );
int _mathtreeDupNoData  ( struct SMATHTREE *, struct SMATHTREE *, unsigned long );
int _mathtreeEvaluate   ( struct SMATHTREE *, unsigned long ); 
int _mathtree1Oper      ( struct SMATHTREE *, struct SMATHTREE *, unsigned long );
int _mathtree2Oper      ( struct SMATHTREE *, struct SMATHTREE *, struct SMATHTREE *, unsigned long );
int _mathtreeFree       ( struct SMATHTREE *, unsigned long );
int _mathtreeFromStr    ( struct SMATHTREE *, const char *, int, unsigned long ); 
int _mathtreeHasSkip      ( struct SMATHTREE *, unsigned long );
int _mathtreeImportData ( struct SMATHTREE *, struct SMATHTREE *, unsigned long );
int _mathtreeImportValStrSub( struct SMATHTREE *, struct SMATHTREEVALSTR *, unsigned long ); 
int _mathtreeInit         ( struct SMATHTREE *, unsigned long ); 
int _mathtreeInspect      ( struct SMATHTREE *, unsigned long ); 
int _mathtreeSetImportSub ( struct SMATHTREE *, struct SMATHTREE *, unsigned long );
int _mathtreeSetSkip      ( struct SMATHTREE *, unsigned long );

struct SMATHTREE *_mathtreeTest( unsigned long ); 
struct SMATHTREE *_mathtreeTestChar( unsigned long ); 

                                      
                                      /* for internal scanning */    
struct SMATHTREESCAN {
   char *pcStr; int ncStr;            /* input string */ 
   int iOpClass;                      /* operation class */
};
                                      /* op classes */  
enum { MATHTREESCAN_CLASSPOWER=0,       /* **, ^ */ 
       MATHTREESCAN_CLASSDOT,         /* ./ */
       MATHTREESCAN_CLASSDASH,        /* +- */
};
    
/*

   _mathtreeScan    ( mt, mtscan, flag )
                    o scans an expression, recursively
                    o to complete test
                      o syntax errors (from state table)
                    o defer function recognition to allow user functions 
                    ? assignment/constant/scalar by psMT->psMT[0] 
                    o calls
                      o _mathtreeInit()
                      o _mathtreeScan()
                       
*/
int _mathtreeScan( struct SMATHTREE *, struct SMATHTREESCAN *, unsigned long ); 


                                      /* error handling ... */
char *_mathtreeStrError( int );
                                      /* error codes */
enum {
   ERRMATHTREE_SYS=1,
   ERRMATHTREE_BADADDR,
   ERRMATHTREE_DIMMISMATCH,
   ERRMATHTREE_TYPEMISMATCH,
   ERRMATHTREE_BADDATATYPE,
   ERRMATHTREE_TOOLARGE,
   ERRMATHTREE_BADFUNCTION,
   ERRMATHTREE_BADSTR,
   ERRMATHTREE_BADSUBSCRIPTRANGE,
   ERRMATHTREE_BADNUMBER,
   ERRMATHTREE_IMBALANCE,      /* 0xB */ 
   ERRMATHTREE_SYNTAX,         /* 0xC */ 
   ERRMATHTREE_BADCHAR,        /* 0xD */ 
   ERRMATHTREE_MAXDIM,
   ERRMATHTREE_BADNARG,
   ERRMATHTREE_BADDIM,
   ERRMATHTREE_BADCOUNT,
                                      /* 18, 0x12 */
   ERRMATHTREE_NODATA,
   ERRMATHTREE_BADINDEXSTEP,
   ERRMATHTREE_BADSTEP,
   ERRMATHTREE_VECTOR2SCALAR,
   ERRMATHTREE_DOMAIN,                /* 0x16 */  
   ERRMATHTREE_BADSUBSCRIPT,
   ERRMATHTREE_BADNAME,
};

#endif
